import org.apache.poi.ss.usermodel.HorizontalAlignment
import org.apache.poi.ss.util.CellReference
import org.apache.poi.xssf.usermodel.XSSFWorkbook
import org.json.JSONObject
import org.jsoup.Jsoup
import org.openqa.selenium.By
import org.openqa.selenium.firefox.FirefoxDriver
import org.openqa.selenium.firefox.FirefoxDriverLogLevel
import org.openqa.selenium.firefox.FirefoxOptions
import org.openqa.selenium.firefox.GeckoDriverService
import java.io.File
import java.io.FileOutputStream
import java.net.URL
import java.time.Instant
import java.time.LocalDateTime
import java.time.ZoneId
import java.time.format.DateTimeFormatter
import java.util.*

private val workbook: XSSFWorkbook = XSSFWorkbook()
private var tableCounter = 0
fun main(vararg args: String) {
    val argumentHolder = ArgumentParser(*args).getArgumentHolder()
    val fileName = argumentHolder.fileName
    println("arguments provided ${argumentHolder}")
    argumentHolder.url.split("||")
        .map {
            argumentHolder.copy(url = it, fileName = URL(it).toURI().path.split("/").last())
        }.forEach { scrapData(it) }
    writeExcelFile(fileName)
}

private fun writeExcelFile(fileName: String) {
    FileOutputStream(File("$fileName.xlsx"), false).use { workbook.write(it) }
    workbook.close()
    println("Data saved in file ${File("$fileName.xlsx").absolutePath}")
}

private fun scrapData(argumentHolder: ArgumentHolder) {
    val link = getChartLink(argumentHolder.url)
    if (link.isNotEmpty()) {
        val queryParams = URL(link).getQueryParams()
        val queryParamsMain = URL(argumentHolder.url).getQueryParams()
        val to = Date().time / 1000
        val from = argumentHolder.fromTime
        val dataUrl =
            "https://tvc4.investing.com/${queryParams["carrier"]}/${queryParams["time"]}/${queryParams["domain_ID"]}/${queryParams["lang_ID"]}/${queryParams["timezone_ID"]}/history?symbol=${queryParamsMain["cid"]?.ifEmpty { "172" } ?: "172"}&resolution=${argumentHolder.resolution}&from=${from}&to=${to}"
        System.setProperty(GeckoDriverService.GECKO_DRIVER_LOG_PROPERTY, "/dev/null")
        GeckoDriverService.createDefaultService().apply {
            start()
        }
        FirefoxDriver(FirefoxOptions().setLogLevel(FirefoxDriverLogLevel.ERROR).apply {
            addArguments("--headless")
        }).apply {
            get(dataUrl)
            saveToExcel(
                getTimeEventData(JSONObject(findElement(By.tagName("body")).text)),
                argumentHolder.zoneId,
                argumentHolder.fileName
            )
            this.close()
        }
    }
}

private fun getTimeEventData(jsonObject: JSONObject): List<TimeEventHolder> {
    val list = mutableListOf<TimeEventHolder>()
    var size: Int
    val t = jsonObject.getJSONArray("t").also { size = it.length() }
    val o = jsonObject.getJSONArray("o")
    val l = jsonObject.getJSONArray("l")
    val h = jsonObject.getJSONArray("h")
    val c = jsonObject.getJSONArray("c")
    for (i in 0 until size) {
        list.add(
            TimeEventHolder(
                t.getLong(i),
                o.getDouble(i),
                c.getDouble(i),
                h.getDouble(i),
                l.getDouble(i),
            )
        )
    }


    return list
}

private fun saveToExcel(list: List<TimeEventHolder>, zoneId: ZoneId, fileName: String) {
    val sheet = workbook.createSheet(fileName)
    val headerStyle = workbook.createCellStyle().apply {
        setFont(workbook.createFont().apply { bold = true })
        alignment = HorizontalAlignment.CENTER
    }
    val header = sheet.createRow(0)
    for ((id, headers) in arrayOf("Date", "Time", "Opening", "Highest", "Lowest", "Closing").withIndex()) {
        header.createCell(id).apply {
            cellStyle = headerStyle
            setCellValue(headers)
        }
    }
    list.forEachIndexed { index, timeEventHolder ->
        sheet.createRow(index + 1).apply {
            Instant.ofEpochSecond(timeEventHolder.time).also {
                LocalDateTime.ofInstant(it, zoneId).also {
                    createCell(0).setCellValue(it.format(DateTimeFormatter.ISO_DATE))
                    createCell(1).setCellValue(it.format(DateTimeFormatter.ISO_TIME))
                }
            }

            createCell(2).setCellValue(timeEventHolder.opening)
            createCell(3).setCellValue(timeEventHolder.highest)
            createCell(4).setCellValue(timeEventHolder.lowest)
            createCell(5).setCellValue(timeEventHolder.closing)
        }
    }
    sheet.createTable(
        workbook.creationHelper.createAreaReference(
            CellReference(sheet.first().first()),
            CellReference(sheet.last().last())
        )
    ).apply {
        name = "Table${++tableCounter}"
        displayName = "Table${++tableCounter}"
        ctTable.addNewAutoFilter().ref = area.formatAsString()
    }
    sheet.first().forEach { sheet.autoSizeColumn(it.columnIndex) }
}

private fun URL.getQueryParams(): Map<String, String> {
    return query?.let {
        if (it.startsWith("&")) it.substring(1)
        else it
    }?.getQueryParam() ?: emptyMap()
}

private fun String.getQueryParam(): Map<String, String> {
    return split("&").map {
        it.indexOf("=").let { index ->
            it.substring(0, index) to it.substring(index + 1)
        }
    }.toMap()
}

private fun getChartLink(url: String): String {
    return Jsoup.connect(url).get()
        .body().getElementsByTag("iframe")
        .last()?.attr("src") ?: ""
}

data class TimeEventHolder(
    val time: Long = 0,
    val opening: Double = 0.0,
    val closing: Double = 0.0,
    val highest: Double = 0.0,
    val lowest: Double = 0.0
)

data class ArgumentHolder(
    var url: String,
    var resolution: String,
    var fromTime: Long,
    var zoneId: ZoneId,
    var fileName: String
)

class ArgumentParser(vararg arguments: String) {
    private var holder = ArgumentHolder(
        "https://www.investing.com/indices/germany-30-chart||https://www.investing.com/indices/us-spx-500-futures-chart?cid=1175153||https://www.investing.com/indices/us-30-futures-chart?cid=1175152||https://www.investing.com/indices/nq-100-futures-chart?cid=1175151",
        "15",
        (Calendar.getInstance().apply { add(Calendar.DAY_OF_MONTH, -15) }.time.time / 1000),
        ZoneId.systemDefault(),
        "dataFile"
    )
    private var validResolutions = mapOf(
        "1m" to ("1" to (Calendar.getInstance().apply { add(Calendar.DAY_OF_MONTH, -2) }.time.time / 1000)),
        "5m" to ("5" to (Calendar.getInstance().apply { add(Calendar.DAY_OF_MONTH, -5) }.time.time / 1000)),
        "15m" to ("15" to (Calendar.getInstance().apply { add(Calendar.DAY_OF_MONTH, -15) }.time.time / 1000)),
        "30m" to ("30" to (Calendar.getInstance().apply { add(Calendar.MONTH, -1) }.time.time / 1000)),
        "45m" to ("15" to (Calendar.getInstance().apply { add(Calendar.MONTH, -1) }.time.time / 1000)),
        "1h" to ("60" to (Calendar.getInstance().apply { add(Calendar.MONTH, -2) }.time.time / 1000)),
        "2h" to ("60" to (Calendar.getInstance().apply { add(Calendar.MONTH, -4) }.time.time / 1000)),
        "4h" to ("60" to (Calendar.getInstance().apply { add(Calendar.MONTH, -3) }.time.time / 1000)),
        "5h" to ("300" to (Calendar.getInstance().apply { add(Calendar.MONTH, -10) }.time.time / 1000)),
        "M" to ("M" to (Calendar.getInstance().apply { add(Calendar.YEAR, -30) }.time.time / 1000)),
        "D" to ("D" to (Calendar.getInstance().apply { add(Calendar.YEAR, -1) }.time.time / 1000)),
        "W" to ("W" to (Calendar.getInstance().apply { add(Calendar.YEAR, -5) }.time.time / 1000))
    )

    init {
        val keys = mutableListOf<String>()
        val values = mutableListOf<String>()
        arguments.forEachIndexed { index, s ->
            if (index % 2 == 0) keys.add(s)
            else values.add(s)
        }
        if (keys.size != values.size) throw IllegalArgumentException("Invalid argument")
        keys.zip(values).forEach {
            if (it.first == "-d") {
                if (validResolutions.containsKey(it.second)) {
                    validResolutions.get(it.second)?.also {
                        holder.resolution = it.first
                        holder.fromTime = it.second
                    }
                } else {
                    println("Illegal resolution setting defaulting to 15 minutes")
                    println("The value should be from following ${validResolutions.keys}")
                }
            } else if (it.first == "-u") {
                if (it.second.trim().isEmpty()) throw IllegalArgumentException("Please at least send one url")
                holder.url = it.second
            } else if (it.first == "-f") {
                if (it.second.trim().isEmpty()) throw IllegalArgumentException("File name cannot be empty")
                holder.fileName = it.second
            } else if (it.first == "-z") {
                runCatching {
                    holder.zoneId = ZoneId.of(it.second)
                }.onFailure {
                    println("wrong zone id it should be like (Asia/Karachi) setting to system default (${ZoneId.systemDefault()})")
                }
            }
        }
    }

    fun getArgumentHolder(): ArgumentHolder {
        return holder
    }
}


//java -jar stackOverFLow.jar -d M -z "Asia/Karachi" -f "wholeDataFile" for all links
//java -jar stackOverFLow.jar -u "https://www.investing.com/indices/nq-100-futures-chart?cid=1175151" -d M -z "Asia/Karachi"
//java -jar stackOverFLow.jar -u "https://www.investing.com/indices/us-30-futures-chart?cid=1175152" -d M -z "Asia/Karachi"
//java -jar stackOverFLow.jar -u "https://www.investing.com/indices/us-spx-500-futures-chart?cid=1175153" -d M -z "Asia/Karachi"
//java -jar stackOverFLow.jar -u "https://www.investing.com/indices/germany-30-chart" -d M -z "Asia/Karachi"
//-u "URL you want to scrap if multiple then split by || " -d "resolution" -z "Zone id for converting time to" -f "file name"
//resolution value can only be among
//here m = minute,  h = hour , M = month, W = week , D = day
//1m
//5m
//15m
//30m
//45m
//1h
//2h
//4h
//5h
//M
//D
//W